#include <io.h>
#include <stdio.h>
#include <windows.h>
#include "mpc\in_mpc.h"
#include "mpc\requant.h"
#include "mpc\huffsv46.h"
#include "mpc\huffsv7.h"
#include "mpc\bitstream.h"
#include "mpc\mpc_dec.h"
#include "mpc\minimax.h"

#define CUEFIX_VERSION  "0.6"
#define CUEFIX_WAVNAME  "Range.wav"

// -------------------------------------

class cue_file {
private:
    size_t lines;
    char   **line;

public:
    cue_file ();
    ~cue_file ();
    int newline ( const char *str );
    int write ( const char *fn );
};

cue_file::cue_file ()
{
    line = NULL;
    lines = 0;
}

cue_file::~cue_file ()
{
    if ( line ) {
        for ( size_t i = 0; i < lines; i++ ) {
            if ( line[i] ) free ( line[i] );
        }
        free ( line );
    }
}

int cue_file::newline ( const char *str )
{
    void *line_b = (void *)line;
    line = (char **)realloc ( line, (lines+1) * sizeof (char *) );
    if ( !line ) {
        line = (char **)line_b;
        return 0;
    }
    line[lines] = strdup ( str );
    if ( !line[lines] ) return 0;
    lines++;
    return 1;
}

int cue_file::write ( const char *fn )
{
    if ( lines == 0 ) return 0;

    FILE *fp = fopen ( fn, "wt" );
    if ( !fp ) return 0;

    for ( size_t i = 0; i < lines; i++ ) {
        if ( fputs (line[i], fp) == EOF ) {
            fclose ( fp );
            remove ( fn );
            return 0;
        }
    }

    fclose ( fp );
    return 1;
}

// -------------------------------------

// return exact number of frames in MPC, this is a bit slow
Int64_t mpc_frames ( const char* fn )
{
    MPC_decoder *theDec = NULL;
    Reader* reader = NULL;
    StreamInfo *Info = NULL;
 
    FILE *fp = fopen ( fn, "rb" );
    if ( !fp ) {
        printf ( "'%s': file open failed\n", fn );
        return 0;
    }

    reader = new Reader_file ( fp );
    Info = new StreamInfo ();
    if ( Info->ReadStreamInfo (reader) != 0 ) {
        printf ( "'%s': unrecognized format\n", fn );
        if ( reader ) delete reader;
        if ( Info ) delete Info;
        return 0;
    }

    theDec = new MPC_decoder ( reader );
    theDec->RESET_Globals   ();
    theDec->RESET_Synthesis ();
    theDec->SetStreamInfo ( Info );
    theDec->FileInit ();

    Int64_t samples = 0;
    bool broken = false;

    while ( !broken ) {
        unsigned int RING = theDec->Zaehler;
        int valid_samples = theDec->DECODE ( NULL );
        theDec->UpdateBuffer ( RING );
        if ( theDec->FrameWasValid == 0 ) broken = true;
        if ( valid_samples <= 0 ) break;
        samples += valid_samples;
    }

    if ( Info   ) delete Info;
    if ( reader ) delete reader;
    if ( theDec ) delete theDec;

    if ( broken ) return 0;

    return (Int64_t)samples * 75 / 44100;
}

static unsigned long Read32 ( FILE *fp )
{
    unsigned char buff[4];
    if ( fread (buff, 1, 4, fp) != 4 ) return -1;
    return (buff[0] | (buff[1] << 8)) | ((unsigned long)(buff[2] | (buff[3] << 8)) << 16);
}

static unsigned short Read16 ( FILE *fp )
{
    unsigned char buff[2];
    if ( fread (buff, 1, 2, fp) != 2 ) return -1;
    return buff[0] | (buff[1] << 8);
}

// return exact number of frames in WAV
Int64_t wav_frames ( const char* fn )
{
    FILE *fp = fopen ( fn, "rb" );
    if ( !fp ) {
        printf ( "'%s': file open failed\n", fn );
        return 0;
    }

    if ( Read32 (fp) != 0x46464952 ) {                  // 4 Byte: check for "RIFF"
        printf ( "'%s': invalid WAV\n", fn );
        fclose ( fp );
        return 0;
    }
    Read32 ( fp );                                      // 4 Byte: chunk size (ignored)
    if ( Read32 (fp) != 0x45564157 ) {                  // 4 Byte: check for "WAVE"
        printf ( "'%s': invalid WAV\n", fn );
        fclose ( fp );
        return 0;
    }
    if ( Read32 (fp) != 0x20746D66 ) {                  // 4 Byte: check for "fmt "
        printf ( "'%s': invalid WAV\n", fn );
        fclose ( fp );
        return 0;
    }
    Read32 (fp);                                        // 4 Byte: read chunk-size (ignored)
    if ( Read16 (fp) != 0x0001 ) {                      // 2 Byte: check for linear PCM
        printf ( "'%s': invalid WAV\n", fn );
        fclose ( fp );
        return 0;
    }
    unsigned int Channels   = Read16 (fp);              // 2 Byte: read no of channels
    unsigned int SampleFreq = Read32 (fp);              // 4 Byte: read sampling frequency
    Read32 (fp);                                        // 4 Byte: read avg. blocksize (fs*channels*bytepersample)
    Read16 (fp);                                        // 2 Byte: read byte-alignment (channels*bytepersample)
    unsigned int BitsPerSample = Read16 (fp);           // 2 Byte: read bits per sample
    unsigned int BytesPerSample = (BitsPerSample + 7) / 8;

    while ( 1 ) {                                       // search for "data"
        if ( feof (fp) ) {
            fclose ( fp );
            return 0;
        }
        if ( Read16 (fp) != 0x6164 ) continue;
        if ( Read16 (fp) == 0x6174 ) break;
    }

    unsigned int PCMBytes = Read32 (fp);                // 4 Byte: no. of byte in file
    if ( feof (fp) ) {
        fclose ( fp );
        return 0;
    }
    fclose ( fp );

    Int64_t samples = PCMBytes >= 0xFFFFFF00 || PCMBytes == 0 || PCMBytes % (Channels * BytesPerSample) != 0  ?  // finally calculate number of samples
                            0xFFFFFFFFFFFFFFFF : PCMBytes / (BytesPerSample * Channels);

    return (Int64_t)samples * 75 / 44100;
}

// return exact number of frames in file
Int64_t get_frames ( const char* fn )
{
    char *ext = strrchr ( fn, '.' );
    if ( !ext ) return 0;

    if ( !stricmp (ext, ".wav") )
        return wav_frames ( fn );
    else if ( !stricmp (ext, ".mpc") || !stricmp (ext, ".mp+") )
        return mpc_frames ( fn );

    return 0;
}

// -------------------------------------

bool exists ( const char *fn )
{
    return (_access (fn, 0) == 0);
}

// get filename from a line in cue
int getfilename ( char *buffer, const char *string )
{
    const char *ptr = string;

    if ( !strncmp (ptr, "FILE ", 5) ) {
        buffer[0] = '\0';
        ptr += 5;
        bool quoted = false;
        if ( *ptr == '"' ) {
            quoted = true;
            ptr++;
        }

        char        temp[_MAX_PATH+4];
        const char  *s = ptr;
        char        *d = temp;

        if ( quoted ) {
            while ( *s && *s != '"' ) *d++ = *s++;      // copy filename to 'temp'
            if ( *s != '"' ) return -1;                 // something is wrong, abort
        } else {
            while ( *s && *s != ' ' ) *d++ = *s++;      // copy filename to 'temp'
            if ( *s != ' ' ) return -1;                 // something is wrong, abort
        }
        *d = '\0';

        char fname[_MAX_FNAME];
        char ext  [_MAX_EXT  ];

        _splitpath ( temp, NULL, NULL, fname, ext );    // remove drive and path
        _makepath ( buffer, NULL, NULL, fname, ".wav" );// check WAV first  (ugly, fix)
        if ( !exists (buffer) ) {                       // MPC second
            _makepath ( buffer, NULL, NULL, fname, ".mpc" );
            if ( !exists (buffer) )                     // MP+ last
                _makepath ( buffer, NULL, NULL, fname, ".mp+" );
        }
    }

    return 0;
}

// convert string to number of frames (format MM:SS:FF)
Int64_t atofr ( const char *a )
{
    unsigned long tf, ts, tm;
    if ( sscanf (a, "%02u:%02u:%02u", &tm, &ts, &tf) != 3 ) return -1;
    return (Int64_t)((Int64_t)tf + ((Int64_t)ts * 75) + ((Int64_t)tm * 60 * 75));
}

// check if cue sheet needs fixing
int cue_fixneeded ( const char *fn )
{
    FILE* fp = fopen ( fn, "rt" );
    if ( !fp ) {
        printf ( "Failed to open '%s'.\n", fn );
        return -1;
    }

    char str[4096];
    bool gapless = true;
    bool noncompliant = false;
    size_t filelines = 0;
    enum cpos_t {
        na    = -1,
        file  = 0,
        track = 1,
        index = 2,
    } cpos;

    cpos = na;
    while ( fgets (str, sizeof (str), fp) != NULL ) {
        if ( !strncmp (str, "FILE ", 5) && strstr ( str, " WAVE\n") ) {
            filelines++;
            cpos = file;
        }
        else if ( strstr (str, "TRACK ") ) {
            cpos = track;
        }
        else if ( strstr (str, "INDEX ") ) {
            if ( strstr (str, "INDEX 00 ") ) gapless = false;
            if ( cpos == file ) noncompliant = true;
            cpos = index;
        }
    }

    fclose ( fp );
    if ( filelines < 1 ) {
        printf ( "File '%s' is not a valid cue sheet.\n", fn );
        return -1;
    }

    if ( noncompliant ) {
        printf ( "File '%s' is in non-compliant format, no fixing needed.\n", fn );
        return -1;
    }

    if ( filelines == 1 ) {
        printf ( "'%s': only one audio file, no fixing needed.\n", fn );
        return 0;
    }

    if ( gapless ) {
        printf ( "'%s': no gaps, no fixing needed.\n", fn );
        return 0;
    }

    return 1;
}

// convert multi wav cue into gap-corrected single wav cue
int cue_fix ( const char *fn, int force )
{
    if ( !force ) {
        int ret = cue_fixneeded ( fn );
        if ( ret < 0 ) return -1;
        if ( ret == 0 ) return 0;
    }

    if ( _access (fn, 2) != 0 ) {
        printf ( "File '%s' is read only.\n", fn );
        return -1;
    }

    cue_file *outcue = new cue_file ();
    if ( !outcue ) {
        printf ( "Memory allocation failed.\n" );
        return -1;
    }

    FILE* ifile = fopen ( fn, "rt" );
    if ( !ifile ) {
        printf ( "Failed to open '%s'.\n", fn );
        delete outcue;
        return -1;
    }

    char str[4096];
    int filelines = 0;
    char filename[_MAX_PATH] = "";
    Int64_t totalframes = 0;
    Int64_t lastframes = 0;
    Int64_t frameoffs  = 0;

    printf ( "Fixing cue '%s'...\n", fn );

    while ( fgets (str, sizeof (str), ifile) != NULL ) {
        char *ptr;

        ptr = str;
        if ( !strncmp (ptr, "FILE ", 5) && strstr (ptr, " WAVE\n") ) {
            frameoffs = 0;
            filelines++;
            if ( getfilename (filename, str) != 0 ) {
                printf ( "Fixing failed, invalid cue.\n" );
                fclose ( ifile );
                delete outcue;
                return 1;
            }
            totalframes += lastframes;
            lastframes = get_frames ( filename );
            if ( lastframes <= 0 ) {
                printf ( "Fixing failed, error with audio files.\n" );
                fclose ( ifile );
                delete outcue;
                return 1;
            }

            if ( filelines > 1 ) continue;

            strcpy ( str, "FILE \"" CUEFIX_WAVNAME "\" WAVE\n" );
        }

        ptr = strstr ( str, "INDEX " );
        if ( ptr != NULL ) {
            ptr += strlen ( "INDEX xx " );

            if ( !strncmp (ptr-3, "00", 2) ) { // gap
                Int64_t f1 = atofr ( ptr );
                if ( f1 < 0 ) {
                    printf ( "Fixing failed, error with INDEX conversion (1)\n" );
                    fclose ( ifile );
                    delete outcue;
                    return 1;
                }

                char temp[1024];
                strcpy ( temp, str );
                char *ptr1 = strstr ( temp, "INDEX " );
                ptr1 += strlen ( "INDEX xx " );

                if ( fgets (str, sizeof (str), ifile) == NULL ) {
                    printf ( "Fixing failed, invalid cue.\n" );
                    fclose ( ifile );
                    delete outcue;
                    return 1;
                }
                ptr = strstr ( str, "INDEX " );
                if ( ptr == NULL ) {
                    printf ( "Fixing failed, invalid cue.\n" );
                    fclose ( ifile );
                    delete outcue;
                    return 1;
                }
                ptr += strlen ( "INDEX xx " );
                Int64_t f2 = atofr ( ptr );
                if ( f2 < 0 ) {
                    printf ( "Fixing failed, error with INDEX conversion (2)\n" );
                    fclose ( ifile );
                    delete outcue;
                    return 1;
                }
                frameoffs = f2 - f1;
                if ( frameoffs < 0 ) {
                    printf ( "Fixing failed, error with INDEX conversion (3)\n" );
                    fclose ( ifile );
                    delete outcue;
                    return 1;
                }

                bool pregap = false;
                Int64_t frames = f1 + totalframes - frameoffs;
                if ( frames < 0 ) {
                    pregap = true;
                    Int64_t pregapf = -frames;
                    frames += pregapf;

                    unsigned long tf = (unsigned long)(pregapf % 75);
                    unsigned long ts = (unsigned long)(pregapf / 75);
                    unsigned long tm = ts / 60;
                    ts = ts % 60;
                    char pregaps[64];
                    sprintf ( pregaps, "    PREGAP %02u:%02u:%02u\n", tm, ts, tf );

                    if ( !outcue->newline (pregaps) ) {
                        printf ( "Fixing failed, insufficient memory.\n" );
                        fclose ( ifile );
                        delete outcue;
                        return 1;
                    }
                }

                if ( frames < 0 ) {
                    printf ( "Fixing failed, error with INDEX conversion (4)\n" );
                    fclose ( ifile );
                    delete outcue;
                    return 1;
                }

                if ( !pregap ) {
                    unsigned long tf = (unsigned long)(frames % 75);
                    unsigned long ts = (unsigned long)(frames / 75);
                    unsigned long tm = ts / 60;
                    ts = ts % 60;
                    sprintf ( ptr1, "%02u:%02u:%02u\n", tm, ts, tf );

                    if ( !outcue->newline (temp) ) {
                        printf ( "Fixing failed, insufficient memory.\n" );
                        fclose ( ifile );
                        delete outcue;
                        return 1;
                    }
                }
            }

            Int64_t tfr = atofr ( ptr );
            if ( tfr < 0 ) {
                printf ( "Fixing failed, error with INDEX conversion (6)\n" );
                fclose ( ifile );
                delete outcue;
                return 1;
            }
            Int64_t frames = tfr + totalframes - frameoffs;
            if ( frames < 0 ) {
                printf ( "Fixing failed, error with INDEX conversion (7)\n" );
                fclose ( ifile );
                delete outcue;
                return 1;
            }

            unsigned long tf = (unsigned long)(frames % 75);
            unsigned long ts = (unsigned long)(frames / 75);
            unsigned long tm = ts / 60;
            ts = ts % 60;
            sprintf ( ptr, "%02u:%02u:%02u\n", tm, ts, tf );
        }

        if ( !outcue->newline (str) ) {
            printf ( "Fixing failed, insufficient memory.\n" );
            fclose ( ifile );
            delete outcue;
            return 1;
        }
    }

    fclose ( ifile );

    char tmp[_MAX_PATH];
    char tempname[_MAX_PATH];
    tmpnam ( tmp );
    if ( tmp[0] == '\\' || tmp[0] == '/' ) {
        strcpy ( tempname, tmp+1 );
    } else {
        strcpy ( tempname, tmp );
    }

    if ( rename (fn, tempname) != 0 ) {
        printf ( "Failed to backup old cue.\n" );
        delete outcue;
        return -1;
    }

    if ( !outcue->write (fn) ) {
        printf ( "Failed to write new cue.\n" );
        rename (tempname, fn);
        delete outcue;
        return -1;
    }

    remove ( tempname );
    delete outcue;

    printf ( "done.\n" );

    return 0;
}

int cue_fix_files ( const char *path, int force )
{
    struct _finddata_t  fd;
    long                hFile;
    int                 errors = 0;

    if ( (hFile = _findfirst (path, &fd)) != -1L ) {
        if ( !(fd.attrib & _A_SUBDIR || fd.attrib & _A_HIDDEN) ) {
            errors += cue_fix ( fd.name, force );
        }

        while ( _findnext (hFile, &fd) == 0 ) {
            if ( !(fd.attrib & _A_SUBDIR || fd.attrib & _A_HIDDEN) ) {
                errors += cue_fix ( fd.name, force );
            }
        }

        _findclose ( hFile );
    } else {
        printf ( "Can't find file '%s'\n", path );
    }

    return errors;
}

int main ( int argc, char **argv )
{
    int i, errors = 0;
    int force = 0;

    if ( argc < 2 ) {
        printf ( "Cuefix v" CUEFIX_VERSION "\n" );
        printf ( "Usage: cuefix [options] <file>\n" );
        printf ( "Possible options:\n" );
        printf ( " --force : modify any cue\n" );
        return 0;
    }

    for ( i = 1; i < argc; i++ ) {
        if ( !strcmp (argv[i], "--force") )
            force = 1;
        else
            errors += cue_fix_files ( argv[i], force );
    }

    return errors;
}
